__author__ = 'Aaron Mussig'
__author_email__ = 'aaronmussig@gmail.com'
__copyright__ = 'Copyright 2023'
__credits__ = ['Aaron Mussig']
__description__ = 'Automatic pre-curation of GTDB trees.'
__email__ = 'aaronmussig@gmail.com'
__license__ = 'GPL3'
__maintainer__ = 'Aaron Mussig'
__maintainer_email__ = 'aaronmussig@gmail.com'
__name__ = 'gtdb_precurate'
__python_requires__ = '>=3.8'
__status__ = 'Production'
__title__ = 'GTDB Precurate'
__url__ = 'https://github.com/Ecogenomics/gtdb_precurate'
__version__ = '1.0.0'



import json
from pathlib import Path

from gtdb_precurate.model.ranks import RANKS


class RedDict:
    __slots__ = ('d', 'p', 'c', 'o', 'f', 'g', 's')

    def __init__(self, d, p, c, o, f, g, s):
        self.d: float = d
        self.p: float = p
        self.c: float = c
        self.o: float = o
        self.f: float = f
        self.g: float = g
        self.s: float = s
        return

    def get_rank(self, rank_str: str):
        """Returns the RED value for the current rank."""
        return getattr(self, rank_str)

    def get_rank_higher(self, rank_str: str):
        """Returns the RED value for the parent rank."""
        parent_rank = RANKS[RANKS.index(rank_str) - 1]
        return self.get_rank(parent_rank)

    def get_rank_threshold(self, rank: str):
        """Compute the halfway point between the current rank and parent."""
        cur_rank = self.get_rank(rank)
        parent_rank = self.get_rank_higher(rank)
        return (cur_rank + parent_rank) / 2.0

    @classmethod
    def load(cls, path: Path):
        with open(path) as f:
            val = json.loads(f.read())
            new = {k[0]: v for k, v in val.items()}
            new['d'] = 0.0
            return cls(**new)



from collections import defaultdict, deque
from pathlib import Path
from typing import Set, Dict, FrozenSet, Iterator

import dendropy

from gtdb_precurate.model.node_label import NodeLabel
from gtdb_precurate.model.ranks import RANKS, TaxString
from gtdb_precurate.util.tree import parse_node_label


class Tree:

    def __init__(self, tree):
        self.data: dendropy.Tree = tree

        # Pre-compute values required
        self.gids: Set[str] = self.get_gids()
        self.node_depths: Dict[int, FrozenSet[dendropy.Node]] = self.get_node_depths()
        self.node_descs: Dict[dendropy.Node, FrozenSet[dendropy.Node]] = self.get_node_descs()
        self.known_taxa: FrozenSet[str] = self.get_tree_taxa()

    def get_gids(self) -> Set[str]:
        out = set()
        for taxon in self.data.taxon_namespace:
            label = taxon.label
            if label.startswith('G'):
                out.add(label.split('|')[0])
        return out

    def get_node_depths(self):
        """Compute the depth of each node in the tree."""
        out = defaultdict(set)
        queue = deque([(self.data.seed_node, 0)])
        while len(queue) > 0:
            cur_node, cur_depth = queue.popleft()
            out[cur_depth].add(cur_node)
            for child_node in cur_node.child_nodes():
                queue.append((child_node, cur_depth + 1))
        return {k: frozenset(v) for k, v in out.items()}

    def get_node_descs(self):
        """Calculate the descendants of each node in the tree."""
        out = defaultdict(set)
        max_depth = max(self.node_depths.keys())
        for i in reversed(range(max_depth + 1)):
            for cur_node in self.node_depths[i]:
                if cur_node.is_leaf():
                    out[cur_node] = set()
                else:
                    for child_node in cur_node.child_node_iter():
                        out[cur_node].update(out[child_node])
                        out[cur_node].add(child_node)
        return {k: frozenset(v) for k, v in out.items()}

    def get_node_ancestors(self, node: dendropy.Node) -> Iterator[dendropy.Node]:
        """Returns all ancestors of the specified node."""
        cur_node = node.parent_node
        while cur_node is not None:
            yield cur_node
            cur_node = cur_node.parent_node

    def get_tree_taxa(self) -> FrozenSet[str]:
        """Generates a set of all taxa observed in the tree."""
        out = set()
        for node in self.data.postorder_node_iter():
            _, _, taxa = parse_node_label(node)
            if taxa is not None:
                for taxon in taxa.split('; '):
                    out.add(taxon)
        return frozenset(out)

    def remove_red_labels(self):
        for node in self.data.postorder_node_iter():
            if node.is_leaf():
                node.taxon.label = node.taxon.label.split('|')[0]
            else:
                node.label = node.label.split('|')[0]
        return

    def infer_gid_taxstring_from_tree(self) -> Dict[dendropy.Node, TaxString]:
        out = dict()
        for leaf_node in self.data.leaf_node_iter():
            if not leaf_node.taxon.label.startswith('G'):
                continue
            tax_string = defaultdict(list)
            for ancestor in self.get_node_ancestors(leaf_node):
                label = NodeLabel(ancestor)
                if label.taxon is not None:
                    for cur_taxon in label.taxon.split('; '):
                        tax_string[cur_taxon[0]].append(cur_taxon)
            out[leaf_node] = TaxString.from_dict(tax_string)
        return out

    def get_ranks_below_node(self, node: dendropy.Node) -> FrozenSet[str]:
        out = set()
        for desc_node in self.node_descs[node]:
            label = NodeLabel(desc_node)
            if label.taxon is not None:
                for taxon in label.taxon.split('; '):
                    out.add(taxon[0])
        return frozenset(out)

    def get_ranks_above_node(self, node: dendropy.Node) -> FrozenSet[str]:
        out = set()
        for ancestor in self.get_node_ancestors(node):
            label = NodeLabel(ancestor)
            if label.taxon is not None:
                for taxon in label.taxon.split('; '):
                    out.add(taxon[0])
        return frozenset(out)

    def add_taxon_to_node(self, node: dendropy.Node, taxon: str):
        label = NodeLabel(node)

        # Sort the new ranks, deal with the case that one may exist
        new_tax_dict = {taxon[0]: taxon}
        if label.taxon is not None:
            for cur_taxon in label.taxon.split('; '):
                cur_rank = cur_taxon[0]
                assert cur_rank not in new_tax_dict
                new_tax_dict[cur_rank] = cur_taxon
        new_tax_string = list()
        for cur_rank, the_taxon in sorted(new_tax_dict.items(), key=lambda x: RANKS.index(x[0])):
            new_tax_string.append(the_taxon)
        new_tax_string = '; '.join(new_tax_string)

        # This is a species label node
        if label.bs is None:
            new_taxon_next_label = f'{new_tax_string}|RED={label.red:.3f}'
        else:
            new_taxon_next_label = f'{label.bs:.1f}:{new_tax_string}|RED={label.red:.3f}'

        node.label = new_taxon_next_label
        return

    @classmethod
    def load(cls, path: Path):
        tree = dendropy.Tree.get(path=path, schema="newick", preserve_underscores=True)
        return cls(tree)

    def write(self, path: Path):
        self.data.write(path=path, schema="newick", suppress_rooting=True)



import logging
from collections import defaultdict

from gtdb_precurate.model.metadata import MetadataFile
from gtdb_precurate.model.node_label import NodeLabel
from gtdb_precurate.model.ranks import RANKS
from gtdb_precurate.model.red_dict import RedDict
from gtdb_precurate.model.tree import Tree


def get_gids_missing_ranks(tree):
    out = dict()
    for leaf_node, tax_string in tree.infer_gid_taxstring_from_tree().items():

        # We want to only subset this to those leaf nodes what are new
        parent_label = NodeLabel(leaf_node.parent_node)
        if parent_label.taxon is not None and 'unresolved' in parent_label.taxon:
            missing_ranks = tax_string.missing_ranks()
            if len(missing_ranks) > 0:
                out[leaf_node] = missing_ranks
    return out


def get_candidate_nodes_for_missing_ranks(tree, d_leaf_to_missing_ranks, min_bootstrap, red):
    out = defaultdict(dict)
    d_failed = defaultdict(set)
    for leaf_node, missing_ranks in d_leaf_to_missing_ranks.items():

        # Additionally, we want to look up the tree to avoid nesting higher ranks
        cur_ancestor_ranks = tree.get_ranks_above_node(leaf_node) - {'s'}
        cur_ancestor_ranks_sorted = sorted(cur_ancestor_ranks, key=lambda x: RANKS.index(x))
        new_rank_must_be_above_idx = RANKS.index(cur_ancestor_ranks_sorted[-1])

        # Duplicate work, but we want to check it from the highest to lowest rank
        for cur_rank in missing_ranks:

            # Check to see if adding this rank would cause a nesting of higher ranks
            if RANKS.index(cur_rank) <= new_rank_must_be_above_idx:
                d_failed[leaf_node].add(cur_rank)
                break

            red_threshold = red.get_rank_threshold(cur_rank)

            for ancestor in tree.get_node_ancestors(leaf_node):

                # Before checking anything relating to RED/bootstraps, we need
                # to make sure that adding a rank at this node would not
                # cause any conflicts
                cur_desc_ranks = tree.get_ranks_below_node(ancestor)
                if cur_rank in cur_desc_ranks:
                    break

                # Otherwise, check RED/bs
                label = NodeLabel(ancestor)

                # No bootstrap value is a dummy node artifact
                bs_supported = label.bs is None or label.bs >= min_bootstrap

                # Calculate the red threshold
                red_supported = label.red >= red_threshold

                # As we will never have this supported, stop going up
                if not red_supported:
                    break

                # Otherwise, cache this candidate node
                if bs_supported and red_supported:
                    out[cur_rank][leaf_node] = ancestor

            # Do a quick check to see if we were able to obtain a candidate
            if out[cur_rank].get(leaf_node) is None:
                d_failed[leaf_node].add(cur_rank)
    return out, d_failed


BAD_PAIRS = [("I", "L"), ("0", "O"), ("1", "I"), ("5", "S"), ("2", "Z")]


def score_candidate_taxon(taxon):
    for a, b in BAD_PAIRS:
        if a in taxon and b in taxon:
            return 1
    return 0


def create_taxon(rank, meta, new_taxa, tree, gids):
    prohibited_taxa = new_taxa.union(tree.known_taxa)
    wgs_available = set()
    for gid in gids:
        wgs = meta.data[gid]
        if wgs is not None:
            wgs_rank = f'{rank}__{wgs}'
            if wgs_rank not in prohibited_taxa:
                wgs_available.add(wgs_rank)

    # No WGS can be used, just use the first genome id
    if len(wgs_available) == 0:
        first_gid = sorted(gids)[0]
        return f'{rank}__{first_gid}'

    # Otherwise, take the "best" sscoring WGS
    taxa_sorted = sorted(wgs_available, key=lambda x: (score_candidate_taxon(x), x))
    return taxa_sorted[0]


def create_ranks(d_rank_to_leaf_candidates, tree, meta):
    # We will now iterate from the highest to lower rank to find shared nodes
    created = set()
    log = logging.getLogger('gtdb_precurate')
    for cur_rank, d_leaf_to_candidate_node in sorted(d_rank_to_leaf_candidates.items(),
                                                     key=lambda x: RANKS.index(x[0])):

        # We need to find the shared nodes
        d_shared_nodes = defaultdict(set)
        for leaf_node, candidate_node in d_leaf_to_candidate_node.items():
            d_shared_nodes[candidate_node].add(leaf_node)

        log.info(
            f'Found {len(d_shared_nodes):,} nodes for node placement at rank {cur_rank}, representing ({len(d_leaf_to_candidate_node):,} genomes)')

        # Now, we need to find the shared nodes that have the same number of
        # leaf nodes as the number of genomes that are missing this rank
        for shared_node, leaf_nodes in d_shared_nodes.items():
            # What are the genome accessions for these leaf nodes?
            leaf_node_gids = {x.taxon.label.split('|')[0] for x in leaf_nodes}

            # Create the taxon and make it off-limits
            new_taxon = create_taxon(cur_rank, meta, created, tree, leaf_node_gids)
            created.add(new_taxon)

            # Update the tree to have this taxon
            tree.add_taxon_to_node(shared_node, new_taxon)

    return frozenset(created)


def create_denovo_clusters(
        tree: Tree,
        meta: MetadataFile,
        red: RedDict,
        min_bootstrap: float
):
    """Creates de novo clusters from the specified tree and metadata."""

    log = logging.getLogger('gtdb_precurate')

    # First, we need to find all genomes that are missing ranks
    d_leaf_to_missing_ranks = get_gids_missing_ranks(tree)
    log.info(f'Found {len(d_leaf_to_missing_ranks):,} unresolved genomes missing one or more ranks')

    # Now, we need to find the candidate nodes for each of these leaf nodes
    d_rank_to_leaf_candidates, d_leaf_to_missing_candidate_rank = get_candidate_nodes_for_missing_ranks(
        tree,
        d_leaf_to_missing_ranks,
        min_bootstrap,
        red
    )
    if len(d_leaf_to_missing_candidate_rank) > 0:
        log.warning(f'Unable to candidate nodes for {len(d_leaf_to_missing_candidate_rank):,} genomes')

    created = create_ranks(d_rank_to_leaf_candidates, tree, meta)

    return created



from typing import Optional

import dendropy


class NodeLabel:
    __slots__ = ('red', 'bs', 'taxon')

    def __init__(self, node: dendropy.Node):
        red, bs, taxon = self.parse_node_label(node)
        self.red: float = red
        self.bs: Optional[float] = bs
        self.taxon: Optional[str] = taxon

    @staticmethod
    def parse_node_label(node: dendropy.Node):
        # Leaf node, return nothing
        if node.is_leaf():
            _, red = node.taxon.label.split('|RED=')
            return float(red), None, None

        # Internal node, may have all three
        else:

            # Extract the RED that will always be present
            prefix, suffix = node.label.split('|RED=')
            red = float(suffix)
            bs = None
            taxon = None

            # Extract either the taxon or the bootstrap + taxon
            if ':' in prefix:
                bs, taxon = prefix.split(':')
                bs = float(bs)
            else:
                try:
                    bs = float(prefix)
                except ValueError:
                    taxon = prefix
            return red, bs, taxon



import logging
import sys
from pathlib import Path


def init_logger(out_dir: Path, debug=False):
    # Logger setup
    logger = logging.getLogger('gtdb_precurate')
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    # Formatter
    formatter = logging.Formatter('[%(asctime)s | %(levelname)s] - %(message)s', '%Y-%m-%d %H:%M:%S')

    # Console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    # File handler
    file_handler = logging.FileHandler(out_dir / 'gtdb_precurate.log')
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    return logger



import os
import re

from setuptools import setup, find_packages


def read_meta():
    """Read each of the keys stored in __init__.py

    Returns
    -------
    dict[str, str]
        A dictionary containing each of the string key value pairs.
    """
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'gtdb_precurate/__init__.py')
    with open(path) as fh:
        hits = re.findall(r'__(\w+)__ ?= ?["\'](.+)["\']\n', fh.read())
    return {k: v for k, v in hits}


def readme():
    with open('README.md') as f:
        return f.read()


meta = read_meta()
setup(
    author=meta['author'],
    author_email=meta['author_email'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    data_files=[("", ["LICENSE"])],
    description=meta['description'],
    entry_points={
        'console_scripts': [
            'gtdb_precurate = gtdb_precurate.__main__:main'
        ]
    },
    install_requires=["dendropy>=4.1.0"],
    license=meta['license'],
    long_description=readme(),
    long_description_content_type='text/markdown',
    maintainer=meta['maintainer'],
    maintainer_email=meta['maintainer_email'],
    name=meta['name'],
    packages=find_packages(),
    python_requires=meta['python_requires'],
    url=meta['url'],
    version=meta['version']
)



import sys
from pathlib import Path

import typer

from gtdb_precurate import __version__
from gtdb_precurate.method.create_denovo import create_denovo_clusters
from gtdb_precurate.model.metadata import MetadataFile
from gtdb_precurate.model.ranks import RANKS
from gtdb_precurate.model.red_dict import RedDict
from gtdb_precurate.model.tree import Tree
from gtdb_precurate.util.logger import init_logger


def main(
        metadata: Path,
        red_dict: Path,
        red_decorated_tree: Path,
        out_directory: Path,
        min_bootstrap: float = 95.0,
        debug: bool = False
):
    # Create the output directory
    out_directory.mkdir(parents=True, exist_ok=True)

    # Initialise the logger
    log = init_logger(out_directory, debug)
    log.info(f'gtdb_precurate v{__version__}')
    log.info(f'gtdb_precurate {" ".join(sys.argv[1:])}')

    # Create the output paths
    if 'red_decorated' in red_decorated_tree.name:
        tree_path_out = out_directory / red_decorated_tree.name.replace('red_decorated', 'gtdb_precurate')
    else:
        tree_path_out = out_directory / f'{red_decorated_tree.stem}.gtdb_precurate{red_decorated_tree.suffix}'
    report_path_out = out_directory / f'{tree_path_out.stem}_report.tsv'

    # Read the RED dictionary
    log.info(f'Loading RED dictionary: {red_dict.name}')
    red = RedDict.load(red_dict)

    log.info(f'Reading tree: {red_decorated_tree.name}')
    tree = Tree.load(red_decorated_tree)
    log.info(f'Found {len(tree.gids):,} genomes (excluding zombies)')

    log.info(f'Loading metadata: {metadata.name}')
    meta = MetadataFile.load(metadata, gids_required=tree.gids)
    log.info(f'Found {len(meta.data):,} corresponding records')

    log.info('Inferring de novo clusters')
    created = create_denovo_clusters(tree, meta, red, min_bootstrap)

    log.info(f'Removing RED labels from the tree')
    tree.remove_red_labels()

    log.info(f'Writing tree to disk: {tree_path_out}')
    tree.write(tree_path_out)

    log.info(f'Writing report to disk: {report_path_out}')
    with report_path_out.open('w') as f:
        for cur_taxon in sorted(created, key=lambda x: RANKS.index(x[0])):
            f.write(f'{cur_taxon}\n')

    log.info('Done.')

    return


if __name__ == '__main__':
    typer.run(main)



from pathlib import Path
from typing import Dict, Optional, Set


class MetadataFile:
    __slots__ = ('data',)

    def __init__(self, data):
        self.data: Dict[str, str] = data

    @classmethod
    def load(cls, path: Path, gids_required: Optional[Set[str]] = None):
        out = dict()
        with open(path) as f:
            header = f.readline().strip().split('\t')
            header = {k: i for i, k in enumerate(header)}
            for line in f.readlines():
                cols = line.strip().split('\t')
                gid = cols[header['formatted_accession']]

                # If the user has requested a subset, skip if not present
                if gids_required and gid not in gids_required:
                    continue

                wgs_formatted = cols[header['ncbi_wgs_formatted']]
                wgs_formatted = None if wgs_formatted == 'none' else wgs_formatted
                out[gid] = wgs_formatted
        return cls(out)



from typing import List, Dict, Tuple

RANKS = ('d', 'p', 'c', 'o', 'f', 'g', 's')


class TaxString:
    __slots__ = ('d', 'p', 'c', 'o', 'f', 'g', 's')

    def __init__(self, d, p, c, o, f, g, s):
        self.d: Tuple[str] = tuple(d)
        self.p: Tuple[str] = tuple(p)
        self.c: Tuple[str] = tuple(c)
        self.o: Tuple[str] = tuple(o)
        self.f: Tuple[str] = tuple(f)
        self.g: Tuple[str] = tuple(g)
        self.s: Tuple[str] = tuple(s)

    def __repr__(self):
        return self.as_string()

    @classmethod
    def from_dict(cls, input_d: Dict[str, List[str]]):
        return cls(
            d=input_d['d'],
            p=input_d['p'],
            c=input_d['c'],
            o=input_d['o'],
            f=input_d['f'],
            g=input_d['g'],
            s=input_d['s'],
        )

    def as_string(self) -> str:
        out = list()
        for rank in RANKS:
            cur_val = getattr(self, rank)
            if len(cur_val) == 0:
                out.append(f'{rank}__')
            else:
                out.extend(getattr(self, rank))
        return '; '.join(out)

    def missing_ranks(self) -> Tuple[str]:
        out = list()
        for rank in RANKS:
            if len(getattr(self, rank)) == 0:
                out.append(rank)
        return tuple(out)



import dendropy


def parse_node_label(node: dendropy.Node):
    bs = None
    taxon = None

    # Leaf node, return nothing
    if node.is_leaf():
        gid, red = node.taxon.label.split('|RED=')
        red = float(red)
        bs = None
        taxon = None
        return red, bs, taxon

    # Dummy node for leaf nodes, extract taxon and return RED
    else:
        label = node.label

        # Extract the RED that will always be present
        prefix, suffix = label.split('|RED=')
        red = float(suffix)

        # Extract either the taxon or the bootstrap + taxon
        if ':' in prefix:
            bs, taxon = prefix.split(':')
            bs = float(bs)
        else:
            try:
                bs = float(prefix)
            except ValueError:
                taxon = prefix

    return red, bs, taxon
